/* Copyright (c) 1997-1999 Miller Puckette.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.txt," in this distribution.

iemlib written by Thomas Musil (c) IEM KUG Graz Austria 2000 */


/* ---------- FIR~ - FIR-filter with table~-coef ----------- */

typedef struct sigFIR
{
    t_object x_obj;
    float    *x_coef_beg;
    float    *x_history_beg;
    int      x_index;
    int      x_npoints;
    t_symbol *x_tab_ref_name;
    float    x_msi;
} t_sigFIR;

t_class *sigFIR_class;

static t_int *sigFIR_perform(t_int *w)
{
    float *in = (float *)(w[1]);
    float *out = (float *)(w[2]);
    t_sigFIR *x = (t_sigFIR *)(w[3]);
    int n = (t_int)(w[4]), i=x->x_index, j, k, np=x->x_npoints;
    float sum=0.0, *coef=x->x_coef_beg;
    float *hist1=x->x_history_beg, *hist2, *hh;

    if(!coef) goto sigfirperfzero;

    hist1 += i;
    hist2 = hist1;
    hist2 += np;
    for(j=0; j<n; j++)
    {
	sum = 0.0;
        coef=x->x_coef_beg;
	*hist1++ = *in;
	*hist2 = *in++;
        hh = hist2;
	for(k=0; k<np; k++)
	{
            sum += *coef++ * (*hh--);
	}
	*out++ = sum;
	hist2++;
        i++;
	if(i >= np)
	{
	    i -= np;
	    hist1 -= np;
            hist2 -= np;
	}
    }
    x->x_index = i;
    return(w+5);

sigfirperfzero:

    while(n--) *out++ = 0.0;
    return(w+5);
}

static t_int *sigFIR_perf8(t_int *w)
{
    float *in = (float *)(w[1]);
    float *out = (float *)(w[2]);
    t_sigFIR *x = (t_sigFIR *)(w[3]);
    int n = (t_int)(w[4]), i=x->x_index, j, k, np=x->x_npoints;
    float sum=0.0, *coef=x->x_coef_beg;
    float *hist1=x->x_history_beg, *hist2, *hh;

    if(!coef) goto sigfirperf8zero;

    hist1 += i;
    hist2 = hist1;
    hist2 += np;
    for(j=0; j<n; j++)
    {
	sum = 0.0;
        coef=x->x_coef_beg;
	*hist1++ = *in;
	*hist2 = *in++;
        hh = hist2-8;
	for(k=0; k<np; k+=8, coef+=8, hh-=8)
	{
	    sum += coef[0] * hh[7];
	    sum += coef[1] * hh[6];
            sum += coef[2] * hh[5];
	    sum += coef[3] * hh[4];
            sum += coef[4] * hh[3];
	    sum += coef[5] * hh[2];
            sum += coef[6] * hh[1];
            sum += coef[7] * hh[0];
	}
	*out++ = sum;
	hist2++;
        i++;
	if(i >= np)
	{
	    i -= np;
	    hist1 -= np;
            hist2 -= np;
	}
    }
    x->x_index = i;
    return(w+5);

sigfirperf8zero:

    while(n--) *out++ = 0.0;
    return(w+5);
}

static void sigFIR_free(t_sigFIR *x)
{
    if(x->x_history_beg)
        freebytes(x->x_history_beg, (16+2*x->x_npoints)*sizeof(float));
}

void sigFIR_set(t_sigFIR *x, t_symbol *ref, t_floatarg np)
{
    t_garray *ga;
    
    x->x_tab_ref_name = ref;
    if(!(ga = (t_garray *)pd_findbyclass(x->x_tab_ref_name, garray_class)))
    {
    	if(*ref->s_name)
    	    error("FIR~: %s: no such table~", x->x_tab_ref_name->s_name);
    	x->x_coef_beg = 0;
    }
    else if(!garray_getfloatarray(ga, &x->x_npoints, &x->x_coef_beg))
    {
    	error("%s: bad template for FIR~", x->x_tab_ref_name->s_name);
    	x->x_coef_beg = 0;
    }
    else garray_usedindsp(ga);
    x->x_index = 0;
    if(!x->x_coef_beg && ((int)np > x->x_npoints))
    {
	sigFIR_free(x);
        x->x_npoints = (int)np;
        x->x_history_beg = (float *)getbytes((16+2*x->x_npoints)*sizeof(float));
    }
}

static void sigFIR_dsp(t_sigFIR *x, t_signal **sp)
{
    sigFIR_set(x, x->x_tab_ref_name, x->x_npoints);
    if((x->x_npoints)&7)
        dsp_add(sigFIR_perform, 4, sp[0]->s_vec, sp[1]->s_vec, x, sp[0]->s_n);
    else	
        dsp_add(sigFIR_perf8, 4, sp[0]->s_vec, sp[1]->s_vec, x, sp[0]->s_n);
}

static void *sigFIR_new(t_symbol *ref, t_floatarg np)
{
    t_sigFIR *x = (t_sigFIR *)pd_new(sigFIR_class);

    outlet_new(&x->x_obj, &s_signal);
    x->x_msi = 0;
    x->x_tab_ref_name = ref;
    x->x_coef_beg = 0;
    if((int)np < 1)
        np = 1.0;
    x->x_npoints = (int)np;
    x->x_history_beg = (float *)getbytes((16+2*x->x_npoints)*sizeof(float));
    x->x_index = 0;
    return(x);
}

void sigFIR_setup(void)
{
    sigFIR_class = class_new(gensym("FIR~"), (t_newmethod)sigFIR_new,
	    (t_method)sigFIR_free, sizeof(t_sigFIR), 0, A_DEFSYM, A_DEFFLOAT, 0);
    CLASS_MAINSIGNALIN(sigFIR_class, t_sigFIR, x_msi);
    class_addmethod(sigFIR_class, (t_method)sigFIR_dsp, gensym("dsp"), 0);
    class_addmethod(sigFIR_class, (t_method)sigFIR_set,
    	gensym("set"), A_SYMBOL, A_FLOAT, 0);
    class_sethelpsymbol(sigFIR_class, gensym("iemhelp/help-FIR~"));
}
